# -*- coding: utf8 -*-
from __future__ import unicode_literals

import logging

import six

from driverng import Worklist
from driverng.astm import (
    AstmMessageV2,
    HeaderSegment,
    HprimAD,
    HprimCE,
    HprimCNA,
    HprimDate,
    HprimFloat,
    HprimForeign,
    HprimList,
    HprimMultiple,
    HprimPN,
    HprimString,
    Segment,
)

logger = logging.getLogger("driverng.hprim")


class HprimSegmentH(HeaderSegment):

    nom_fichier = HprimString(3, max_length=12)
    emetteur = HprimMultiple(
        5, code=HprimString(1, obligatoire=True), nom=HprimString(2, obligatoire=True)
    )

    adresse = HprimAD(6)

    # ORM: Demande
    # ORU: Resultat
    # ADM: Admission
    # FAC: Facturation
    # REG: Reglement
    # ERR: Erreur
    type_message = HprimString(7, obligatoire=True, max_length=7)
    telephone = HprimList(8)

    recepteur = HprimMultiple(
        10, code=HprimString(1, obligatoire=True), nom=HprimString(2, obligatoire=True)
    )

    # P: Production
    # T: Test
    # D: Debug
    mode_traitement = HprimString(12, max_length=1, default="P")
    version = HprimString((13, 1))
    type_liaison = HprimString((13, 2), default="L")
    date = HprimDate(14, obligatoire=True)

    patients = HprimForeign("P", "header")
    end = HprimForeign("L", "header", one2one=True, required=True)


class HprimSegmentP(Segment):

    patient_demandeur = HprimString((3, 1), max_length=16)
    patient_fusion = HprimString((3, 2))
    patient_executant = HprimString((4, 1), max_length=16)
    no_adm_cli = HprimString((5, 1))
    rang_naissance = HprimString((5, 2), max_length=2)
    nom = HprimPN(6, max_length=48)
    nomjf = HprimString(7, max_length=24)
    birth_identity = HprimList(
        7,
        basefield=HprimMultiple(
            0,
            last_name=HprimString(1),
            first_name=HprimString(2),
            further_names=HprimString(3),
        ),
    )
    date_nai = HprimDate(8, fmt_date="%Y%m%d")
    # M/MS/F/FS/U
    sexe = HprimString(9)
    animal_breed_code = HprimString(10)
    adresse = HprimAD(11)
    ins_number = HprimList(
        12,
        basefield=HprimMultiple(
            0,
            identifier=HprimString(1),
            type=HprimString(2),
            call_date=HprimDate(3),
            effective_date=HprimDate(4),
            end_date=HprimDate(5),
            oid=HprimString(6),
        ),
    )
    telephone = HprimList(13)
    medecins = HprimList(14, basefield=HprimCNA(0))
    traite_lc1 = HprimString(15)
    traite_lc2 = HprimString(16)
    taille = HprimString(17)
    poids = HprimFloat(18)
    diag_pat = HprimString(19)
    diagnostic_patient = HprimList(19, basefield=HprimCE(0))
    trait_medical = HprimString(20)
    traitement_medical = HprimList(20)
    regime = HprimString(21)
    comment_dem1 = HprimString((22, 1))
    comment_dem1_libelle = HprimString((22, 2))
    comment_dem2 = HprimString(23)
    date_entree = HprimDate((24, 1), fmt_date="%Y%m%d")
    date_sortie = HprimDate((24, 2), fmt_date="%Y%m%d")
    stat_adm = HprimString(25, max_length=2)
    localisation = HprimMultiple(
        26,
        max_length=100,
        lit=HprimString(1),
        chambre=HprimString(2),
        corr=HprimString(3),
    )
    class_diag = HprimString(27)
    situ_maritale = HprimString(29)
    precaution = HprimString(30)
    langue = HprimString(31)
    stat_conf = HprimString(32)
    date_modif = HprimDate(33, fmt_date="%Y%m%d")
    date_deces = HprimDate(34, fmt_date="%Y%m%d")
    birth_place = HprimList(
        35,
        basefield=HprimMultiple(
            0,
            name=HprimString(1),
            insee_code=HprimString(2),
            country=HprimString(3),
            postal_code=HprimString(4),
        ),
    )
    identity_obtaining_mode = HprimString(36)
    identity_document = HprimString(37)
    birthdate_modified = HprimString(38)
    identity_status = HprimString(39)
    identity_attribut = HprimString(40)
    lunar_date = HprimList(
        41,
        basefield=HprimMultiple(
            0, day=HprimString(1), month=HprimString(2), year=HprimString(3)
        ),
    )
    patient_mail = HprimString(42)

    orders = HprimForeign("OBR", "order")


class HprimSegmentOBR(Segment):

    tube_demandeur = HprimString((3, 1))
    no_demandeur = HprimString((3, 2))
    tube_executant = HprimString((4, 1))
    no_executant = HprimString((4, 2))
    prescription = HprimList(5, basefield=HprimCE(0), obligatoire=True)

    # S: Très urgent
    # A: Urgent
    # R: Routine
    # C: Rappel
    # P: Pré-opératoire
    priorite = HprimString((6, 1))
    date_demande = HprimDate(7)
    date_prel = HprimDate(8)
    date_fin_prel = HprimDate(9)
    volume = HprimString(10, max_length=10)
    preleveur = HprimCNA(11, max_length=60)

    # A: Ajout
    # C: Annulation
    # N: Nouvelle demande
    # L: Nouvelle demande avec prélèvement à effectuer par l'éxecutant
    # R: Modification de demande
    action = HprimString(12)
    risque = HprimString(13, max_length=60)
    renseignement_clinique = HprimString(14, max_length=300)
    date_recu = HprimDate(15)
    nature = HprimMultiple(
        16,
        max_length=300,
        type_tube=HprimMultiple(
            1, code=HprimString(1), label=HprimString(2), codification=HprimString(3)
        ),
        additif=HprimString(2),
        methode_prelevement=HprimString(3),
    )
    prescripteur = HprimList(
        17,
        basefield=HprimMultiple(
            0,
            code_medecin=HprimString(1),
            medecin=HprimPN(2),
            type_medecin=HprimString(3),
            code_service=HprimString(4),
            service=HprimPN(5),
            type_service=HprimString(6),
            # Legacy attributes
            nom_medecin=HprimString(2),
            nom_service=HprimString(5),
        ),
    )
    tel_prescrip = HprimList(18)
    lib1_dem = HprimString(19, max_length=60)
    lib2_dem = HprimString(20, max_length=60)
    lib1_exe = HprimString(21, max_length=60)
    lib2_exe = HprimString(22, max_length=60)
    date_resu = HprimDate(23, fmt_date="%Y%m%d%H%M%S")
    service_executant = HprimMultiple(25, code=HprimString(1), label=HprimString(2))
    stat_ana = HprimString(26, max_length=1)
    dest_copie = HprimList(29)
    mod_trans_pat = HprimString(31, max_length=20)
    valideur = HprimCNA(33)
    assistant = HprimCNA(34)
    technicien = HprimCNA(35)
    ope_saisie = HprimCNA(36)
    date_prv_rendu = HprimDate(37)

    results = HprimForeign("OBX", "order")

    def get_action(self):
        return {"C": "del", "A": "add"}.get(self.action, "new")

    def get_services(self):
        if self.prescripteur and self.prescripteur[0].code_service:
            prescripteur = self.prescripteur[0]
            return [
                {"code": prescripteur.code_service, "name": prescripteur.service.nom}
            ]
        # Si le champ OBR.17.4 est vide, on suppose que le SIL nous met
        # l'information dans le champ OBR.25.1
        elif self.service_executant.code:
            return [
                {
                    "code": self.service_executant.code,
                    "name": self.service_executant.label,
                }
            ]
        else:
            return []

    def get_comments(self, option_identification_demandeur=False):
        return [
            e
            for e in [
                self.renseignement_clinique,
                self.lib1_dem if not option_identification_demandeur else "",
                self.lib2_dem,
            ]
            if e
        ]


class HprimSegmentOBX(Segment):

    type_resu = HprimString(3, max_length=3)
    code = HprimString((4, 1), max_length=22)
    libelle = HprimString((4, 2), max_length=60)
    codification = HprimString((4, 3))
    code2 = HprimString((4, 4), max_length=22)
    libelle2 = HprimString((4, 5), max_length=60)
    codification2 = HprimString((4, 6))
    sous_test = HprimString(5, max_length=20)
    resu = HprimList(6, basefield=HprimCE(0))
    unite = HprimCE(7, max_length=20)
    normale = HprimString((8, 1), max_length=60)
    indic_norm = HprimList(9, max_length=10)
    probabilite = HprimString(10, max_length=5)
    nature_norm = HprimString((11, 1))
    stat_resu = HprimString(12, max_length=2)
    date_alt = HprimDate(13)
    droit_acces = HprimString(14, max_length=20)
    date = HprimDate(15)
    secteur_tech = HprimMultiple(
        16,
        libelle=HprimString(1),
        rang=HprimString(2),
        code_batterie=HprimString(3),
        rang_batterie=HprimString(4),
        rang_test=HprimString(5),
        text=HprimString(6),
    )
    valideur = HprimString(17, max_length=60)
    valideur_code = HprimString((17, 1), max_length=60)
    valideur_nom = HprimString((17, 2), max_length=60)

    # Si ces deux champs sont renseignés, le graphe est à pas fixe et le champ
    # 10.6 contient uniquement les ordonnées des points
    gc_pas = HprimString((18, 3))
    gc_origine = HprimString((18, 4))

    # HORS NORME !!
    analyseur = HprimString(19)
    resultat_controle = HprimString(20)
    # Information demande antérieure Clarilab
    no_demande = HprimString((21, 1))
    date_demande = HprimDate((21, 2))


class HprimSegmentL(Segment):

    nombre_segment_patient = HprimString(4, max_length=4)
    nombre_segment = HprimString(5, max_length=10)
    numero_lot = HprimString(6, max_length=12)


class HprimSegmentC(Segment):
    # P => Émis par le demandeur
    # L => Remis par l'éxecutant
    origine = HprimString(3, max_length=1)
    texte = HprimList(4)


class Driver(AstmMessageV2):
    SPEC = {"code": "hprim", "label": "Hprim 2.1"}
    DEFAULT_SEPARATORS = "|~^\\&"

    SEGMENT_CLASS_MAP = {
        "H": HprimSegmentH,
        "P": HprimSegmentP,
        "OBR": HprimSegmentOBR,
        "OBX": HprimSegmentOBX,
        "L": HprimSegmentL,
        "C": HprimSegmentC,
    }
    ADDENDUM_SEGMENT_SIZE = 220

    OPTIONS = [
        AstmMessageV2.Option(
            "hprim_code_emetteur", label="Code émetteur", default="MCA"
        ),
        AstmMessageV2.Option("hprim_nom_emetteur", label="Nom émetteur"),
        AstmMessageV2.Option("hprim_code_recepteur", label="Code récepteur"),
        AstmMessageV2.Option("hprim_nom_recepteur", label="Nom récepteur"),
        AstmMessageV2.BooleanOption(
            "option_comments_in_flags",
            label="Intégrer les commentaires OBX en alarme",
            default=False,
        ),
        AstmMessageV2.BooleanOption(
            "option_envoi_technicien",
            label="Envoyer le code technicien transmis par l'automate au lieu du code du valideur en cas de validation automatique",
            default=False,
        ),
        AstmMessageV2.BooleanOption(
            "option_envoi_sample_id",
            label="Multi SIL (OBR.9.3.2=No tube)",
            default=False,
        ),
        AstmMessageV2.BooleanOption(
            "option_envoi_commentaire_patient",
            label="Envoyer le commentaire patient avec l'envoi de la demande",
            default=False,
        ),
        AstmMessageV2.BooleanOption(
            "option_ipp_executant",
            label="Renvoyer l'IPP du patient dans le champs 8.4",
            default=False,
        ),
        AstmMessageV2.BooleanOption(
            "option_envoyer_fic",
            label="Envoyer les pièces jointes de la demande dans des segments OBX",
            default=False,
        ),
        AstmMessageV2.BooleanOption(
            "option_identification_demandeur",
            label="Identification demandeur OBR.9.19",
            default=False,
        ),
        AstmMessageV2.IntegerOption(
            "option_annulation_sans_ipp_age_max",
            label="Accepter les demandes d'annulation sans IPP ni OBR 9.3.2. Préciser l'age maximum de la demande (en jour).",
            default=0,
        ),
        AstmMessageV2.BooleanOption(
            "option_envoi_code_groupe",
            label="Envoyer le code groupe du demandeur dans le segment H.5.1",
            default=False,
        ),
        AstmMessageV2.BooleanOption(
            "option_bioserveur",
            label="Active les spécificités bioserveur",
            default=False,
        ),
        AstmMessageV2.BooleanOption(
            "option_hprim_integre_rack",
            label="Intégrer le rack du tube présent en résultat accueil (Sysmex XN)",
            default=False,
        ),
        AstmMessageV2.Option(
            "option_rechercher_tube_pour_list_code_resultat",
            label="Liste des codes resultats pour lesquels rechercher le Tube correspondant dans la demande (séparer par un point-virgule) (GENBIO Hexalis)",
        ),
        AstmMessageV2.Option(
            "option_prefixe_code_executant_labo",
            label="Paramétrer un code servant à préfixer le code de l'exécutant laboratoire (connexion SIL) et de son code hprim",
        ),
        AstmMessageV2.BooleanOption(
            "hprim_option_prescripteur",
            label="Traiter le prescripteur dans le champ OBR.9.17",
            default=False,
        ),
        AstmMessageV2.BooleanOption(
            "hprim_concatener_resultats_multiples",
            label='Concaténer les résultats multiples de même code test dans la réponse automate séparés par ";" pour les connexions Hprim/BacExpress de type automate',
            default=False,
        ),
        AstmMessageV2.BooleanOption(
            "hprim_repetiteur_commentaire",
            label="Envoyer les commentaires codés séparés par un répétiteur dans le champ OBX.6",
            default=False,
        ),
        AstmMessageV2.BooleanOption(
            "hprim_send_validator_code_and_name",
            label="Envoyer le code, le nom et le prénom du valideur selon le format CE",
            default=False,
        ),
        AstmMessageV2.BooleanOption(
            "hprim_send_validator_auto",
            label="Envoyer le valideur auto vers le SIL dans le champ OBX.17",
            default=False,
        ),
        AstmMessageV2.BooleanOption(
            "hprim_orm_poccelerator",
            label="Envoyer un message de type ORM sans date de résultat au SIL",
            default=False,
        ),
        AstmMessageV2.BooleanOption(
            "hprim_date_prel",
            label="Traiter la date de prélèvement OBR.9.8 comme date de la demande",
            default=False,
        ),
        AstmMessageV2.Option(
            "hprim_version", label="Version du protocole Hprim", default="H2.1"
        ),
    ]

    def __init__(self, **kwargs):
        self.code_emetteur = kwargs.pop("hprim_code_emetteur", "MCA")
        self.nom_emetteur = kwargs.pop("hprim_nom_emetteur", "")
        self.code_recepteur = kwargs.pop("hprim_code_recepteur", "")
        self.nom_recepteur = kwargs.pop("hprim_nom_recepteur", "")
        self.option_comments_in_flags = kwargs.pop("option_comments_in_flags", False)
        self.option_envoi_technicien = kwargs.pop("option_envoi_technicien", False)
        self.option_envoi_sample_id = kwargs.pop("option_envoi_sample_id", False)
        self.option_ipp_executant = kwargs.pop("option_ipp_executant", False)
        self.option_envoyer_fic = kwargs.pop("option_envoyer_fic", False)
        self.option_identification_demandeur = kwargs.pop(
            "option_identification_demandeur", False
        )
        self.option_annulation_sans_ipp_age_max = kwargs.pop(
            "option_annulation_sans_ipp_age_max", 0
        )
        self.option_envoi_commentaire_patient = kwargs.pop(
            "option_envoi_commentaire_patient", False
        )
        self.option_envoi_code_groupe = kwargs.pop("option_envoi_code_groupe", False)
        self.option_hprim_integre_rack = kwargs.pop("option_hprim_integre_rack", False)
        self.option_bioserveur = kwargs.pop("option_bioserveur", False)
        self.option_prefixe_code_executant_labo = kwargs.pop(
            "option_prefixe_code_executant_labo", ""
        )
        self.hprim_option_prescripteur = kwargs.pop("hprim_option_prescripteur", False)
        self.hprim_concatener_resultats_multiples = kwargs.pop(
            "hprim_concatener_resultats_multiples", False
        )
        self.hprim_repetiteur_commentaire = kwargs.pop(
            "hprim_repetiteur_commentaire", False
        )
        self.hprim_send_validator_code_and_name = kwargs.pop(
            "hprim_send_validator_code_and_name", False
        )
        self.hprim_send_validator_auto = kwargs.pop("hprim_send_validator_auto", False)
        self.hprim_orm_poccelerator = kwargs.pop("hprim_orm_poccelerator", False)
        self.hprim_date_prel = kwargs.pop("hprim_date_prel", False)
        self.hprim_version = kwargs.pop("hprim_version", "H2.1")
        super(Driver, self).__init__(**kwargs)

    def to_string(self, message, *args, **kwargs):
        message.end.nombre_segment_patient = len(message.patients)
        message.end.nombre_segment = len(message.get_segments())
        return super(Driver, self).to_string(message, *args, **kwargs)

    def create(self, **kwargs):
        message = self.SEGMENT_CLASS_MAP["H"](type_segment="H", **kwargs)
        message.end = self.SEGMENT_CLASS_MAP["L"](type_segment="L")
        return message

    def encode_worklist(self, worklist, context):
        if worklist.context not in ("result", "order"):
            raise NotImplementedError("context %s" % (worklist.context,))

        if self.hprim_orm_poccelerator:
            type_message = "ORM"
        elif worklist.context == "result":
            type_message = "ORU"
        else:
            type_message = "ORM"

        msg = self.create(
            date=context["now"],
            type_message=type_message,
            nom_fichier=context.get("filename", worklist.filename),
            version=self.hprim_version,
        )
        msg.emetteur.code = worklist.sender_id or self.code_emetteur
        msg.emetteur.nom = worklist.sender_name or self.nom_emetteur
        msg.recepteur.code = worklist.receiver_id or self.code_recepteur
        msg.recepteur.nom = worklist.receiver_name or self.nom_recepteur

        for patient in worklist.get_astm_worklist(
            unique_together=["action", "priority"],
            action={"del": "C", "add": "A", "new": "N"},
        ):
            if worklist.context == "result" and patient.lab_patient_id:
                patient_demandeur = patient.lab_patient_id
            else:
                patient_demandeur = patient.patient_id

            seg_p = msg.patients.add(
                no_adm_cli=patient.alternate_patient_id,
                date_nai=patient.birthdate,
                sexe=patient.sex,
                date_entree=patient.admission_date,
                date_sortie=patient.admission_end_date,
                poids=patient.weight,
                taille=patient.height,
                rang_naissance=patient.birth_row,
            )
            seg_p.date_deces = patient.deathdate
            if self.option_ipp_executant:
                # Clarilab
                # Exemple 1 : patient.lab_patient_id = 'ABC', patient.patient_id = '123'
                # P|1|ABC|123|
                # Exemple 2 : patient.lab_patient_id = None, patient.patient_id = '123'
                # P|1||123|
                seg_p.patient_demandeur = patient.lab_patient_id
                seg_p.patient_executant = patient.patient_id
            else:
                seg_p.patient_demandeur = patient_demandeur

            if patient.phone:
                if patient.phone == "T":
                    # Spécifique Clarilab
                    seg_p.set_raw("telephone", "~T")
                else:
                    seg_p.telephone = [p for p in patient.phone.split(";") if p]
            if patient.medications:
                seg_p.traitement_medical = patient.medications
            if patient.diagnostic:
                seg_p.diag_pat = patient.diagnostic
            if patient.used_last_name:
                seg_p.nom.nom = patient.used_last_name
            else:
                seg_p.nom.nom = patient.last_name
            if patient.used_first_name:
                seg_p.nom.prenom = patient.used_first_name
            else:
                seg_p.nom.prenom = patient.first_name
            if self.hprim_version == "H2.5" and patient.ins_number:
                birth_identity = seg_p.birth_identity.add()
                birth_identity.last_name = patient.last_name
                birth_identity.first_name = patient.first_name
                further_names = ""
                if patient.further_names:
                    further_names = " ".join(
                        str(element) for element in patient.further_names
                    )
                birth_identity.further_names = further_names
                ins = seg_p.ins_number.add()
                ins.identifier = patient.ins_number
                ins.type = patient.ins_type
                ins.call_date = patient.insi_query_date
                ins.effective_date = patient.ins_effective_date
                ins.end_date = patient.ins_end_date
                ins.oid = patient.oid
            else:
                seg_p.nomjf = patient.maiden_name
            seg_p.nom.civilite = patient.title
            seg_p.stat_adm = patient.admission_status
            seg_p.localisation.lit = patient.bed
            seg_p.localisation.chambre = patient.room
            seg_p.localisation.corr = patient.location
            seg_p.adresse.ville = patient.city
            seg_p.adresse.pays = patient.country
            seg_p.adresse.code_postal = patient.postal_code
            if patient.address:
                if "~" in patient.address:
                    seg_p.adresse.ligne1, seg_p.adresse.ligne2 = patient.address.split(
                        "~", 1
                    )
                else:
                    seg_p.adresse.ligne1 = patient.address
            # envoi du commentaire patient si l'option
            # "option_envoi_commentaire_patient" est activée
            if self.option_envoi_commentaire_patient:
                comments_patient = []
                if patient.patient_comment1:
                    comments_patient.append(patient.patient_comment1)
                if patient.patient_comment2:
                    comments_patient.append(patient.patient_comment2)
                if comments_patient:
                    seg_p.commentaires.add(texte=comments_patient)
            for physician in patient.physicians:
                if physician.type == "attending":
                    med = seg_p.medecins.add()
                    med.code = physician.code
                    med.nom.nom = physician.last_name
                    med.nom.prenom = physician.first_name
                    med.type = physician.codification
                    med.nom.civilite = physician.title
                    med.nom.sobriquet = physician.alias
            if patient.birthplace:
                birth_place = seg_p.birth_place.add()
                birth_place.name = patient.birthplace
                birth_place.insee_code = patient.birthplace_code
                birth_place.country = patient.birthplace_country
                birth_place.postal_code = patient.birthplace_postal_code
            seg_p.identity_obtaining_mode = patient.identity_obtaining_mode
            seg_p.identity_document = patient.identity_document
            seg_p.birthdate_modified = patient.birthdate_modified
            seg_p.identity_status = patient.identity_status
            seg_p.identity_attribut = patient.identity_attribut
            if patient.lunar_date:
                lunar_date = seg_p.lunar_date.add()
                lunar_date.day = patient.lunar_date[6:8]
                lunar_date.month = patient.lunar_date[4:6]
                lunar_date.year = patient.lunar_date[0:4]
            seg_p.patient_mail = patient.email
            self._add_custom_info_patient(seg_p, patient)
            for sample, action, priority in patient.samples:
                if not msg.emetteur.code:
                    msg.emetteur.code = sample.lab
                seg_o = seg_p.orders.add(
                    tube_demandeur=sample.sample_id,
                    no_demandeur=(
                        sample.sample_id
                        if self.option_envoi_sample_id
                        else sample.request_id or sample.sample_id
                    ),
                    priorite=priority or "R",
                    action=action or "N",
                    date_demande=sample.request_date,
                    date_prel=sample.collection_date,
                    date_fin_prel=sample.collection_end_date,
                )
                # Status des résultats ou des actes sur l'ensemble de la demande
                # Utilisé par CLARILAB
                if sample.results_status:
                    seg_o.stat_ana = sample.results_status
                if worklist.context == "result":
                    if self.option_bioserveur:
                        # Bioserveur limite la réception du tube executant
                        # et du numéro executant à 23 carractères.
                        # Ils ignorent le "tube executant", donc on ne le précise pas.
                        seg_o.tube_demandeur = ""
                        seg_o.tube_executant = ""
                        seg_o.no_executant = seg_o.no_demandeur
                    else:
                        seg_o.tube_executant = sample.exe_sample_id or sample.sample_id
                        seg_o.no_executant = (
                            sample.exe_request_id
                            or sample.exe_sample_id
                            or sample.request_id
                            or sample.sample_id
                        )
                if sample.services or sample.physicians:
                    presc = seg_o.prescripteur.add()
                    for service in sample.services:
                        if service:
                            presc.code_service = service.code
                            presc.service.nom = service.name
                    for physician in sample.physicians:
                        if physician.type == "prescriber":
                            presc.code_medecin = physician.code
                            presc.medecin.nom = physician.last_name
                            if physician.phone:
                                seg_o.tel_prescrip = physician.phone.split(";")
                        elif physician.type == "sampler":
                            seg_o.preleveur.code = physician.code
                            seg_o.preleveur.nom.nom = physician.last_name
                            seg_o.preleveur.nom.prenom = physician.first_name
                            seg_o.preleveur.type = physician.codification
                            seg_o.preleveur.nom.civilite = physician.title
                            seg_o.preleveur.nom.sobriquet = physician.alias
                # Natures de l'échantillon
                if sample.nature.code:
                    seg_o.nature.type_tube.code = sample.nature.code
                    seg_o.nature.type_tube.label = sample.nature.label
                    seg_o.nature.type_tube.codification = sample.nature.codification
                    seg_o.nature.additif = sample.nature.additive

                if sample.transcriptionist:
                    seg_o.ope_saisie.code = sample.transcriptionist
                combined_orders_results = []
                for o in sample.orders:
                    if o.code in [x["code"] for x in combined_orders_results]:
                        logger.debug("code %s already present: do not add.", o.code)
                        continue
                    combined_orders_results.append(
                        {
                            "code": o.code,
                            "label": o.label,
                            "codification": o.codification,
                            "alt_code": o.alt_code,
                            "alt_label": o.alt_label,
                            "alt_codification": o.alt_codification,
                        }
                    )
                for r in sample.results:
                    if r.code in [x["code"] for x in combined_orders_results]:
                        logger.debug("code %s already present: do not add.", r.code)
                        continue
                    if r.analyse_code_clarilab and r.analyse_code_clarilab in [
                        x["code"] for x in combined_orders_results
                    ]:
                        logger.debug(
                            "code %s already present: do not add.",
                            r.analyse_code_clarilab,
                        )
                        continue
                    combined_orders_results.append(
                        {
                            "code": (
                                r.analyse_code_clarilab
                                if r.analyse_code_clarilab
                                else r.code
                            ),
                            "label": (
                                r.analyse_label_clarilab
                                if r.analyse_label_clarilab
                                else r.label
                            ),
                            "codification": r.codification,
                            "alt_code": r.alt_code,
                            "alt_label": r.alt_label,
                            "alt_codification": r.alt_codification,
                        }
                    )
                for orderdict in combined_orders_results:
                    test = seg_o.prescription.add()
                    test.code1 = orderdict["code"]
                    if orderdict["label"]:
                        test.label1 = orderdict["label"]
                    if orderdict["codification"]:
                        test.codification1 = orderdict["codification"]
                    if orderdict["alt_code"]:
                        test.code2 = orderdict["alt_code"]
                    if orderdict["alt_label"]:
                        test.label2 = orderdict["alt_label"]
                    if orderdict["alt_codification"]:
                        test.codification2 = orderdict["alt_codification"]
                for result in sample.results:
                    if self.hprim_orm_poccelerator or (
                        worklist.context == "order" and not result.anterior
                    ):
                        date = None
                    else:
                        date = result.date
                    if self.hprim_send_validator_auto and result.validator in (
                        None,
                        "",
                        "auto",
                    ):
                        valideur = "auto"
                        valideur_code = "auto"
                    else:
                        valideur = result.validator
                        valideur_code = result.validator_code
                    if self.option_envoi_technicien and result.operator:
                        valideur = result.operator

                    def do_seg_r():
                        seg_r = seg_o.results.add(
                            type_resu=result.type,
                            code=result.code,
                            libelle=result.label,
                            codification=result.codification,
                            code2=result.alt_code,
                            libelle2=result.alt_label,
                            codification2=result.alt_codification,
                            sous_test=result.sub_code,
                            date=date,
                            indic_norm=result.indic_norm,
                            stat_resu=result.status,
                            analyseur=result.instrument,
                            resultat_controle=result.resultat_controle,
                            normale=result.normality,
                        )
                        if self.hprim_send_validator_code_and_name:
                            seg_r.valideur_nom = result.validator
                            seg_r.valideur_code = valideur_code
                        else:
                            seg_r.valideur = valideur
                        if result.chapter and result.chapter.code:
                            seg_r.secteur_tech.libelle = result.chapter.code
                            seg_r.secteur_tech.rang = result.chapter.position
                        if result.battery and result.battery.code:
                            seg_r.secteur_tech.code_batterie = result.battery.code
                            seg_r.secteur_tech.rang_batterie = (
                                result.battery.chapter_position
                            )
                            seg_r.secteur_tech.rang_test = result.battery.position

                        if result.unit:
                            seg_r.unite.code1 = result.unit
                            seg_r.unite.label1 = result.unit_label
                            seg_r.unite.codification1 = result.unit_codification
                        if result.alt_unit:
                            seg_r.unite.code2 = result.alt_unit
                            seg_r.unite.label2 = result.alt_unit_label
                            seg_r.unite.codification2 = result.alt_unit_codification
                        if self.option_bioserveur:
                            # Bioserveur nécessite de recevoir le code examen
                            # dans le secteur technique (code batterie)
                            seg_r.secteur_tech.code_batterie = result.code
                        return seg_r

                    embended_file = None
                    fic_file = False
                    if result.type == "FIC":
                        fic_file = True
                    if isinstance(result.value, six.string_types):
                        values = [{"value": v} for v in result.value.split(";")]
                    elif isinstance(result.value, dict):
                        values = [result.value]
                    elif (
                        isinstance(result.value, list) and "Base64" not in result.value
                    ):
                        values = result.value
                    elif isinstance(result.value, list) and "Base64" in result.value:
                        embended_file = result.value
                    else:
                        raise NotImplementedError

                    seg_r = None

                    multiple_clarilab = None
                    if values and len(values[0].get("antibiogram", [])):
                        multiple_clarilab = []
                        for a in values[0]["antibiogram"]:
                            if a.get("cutoff") or a.get("diametre"):
                                multiple_clarilab.append(
                                    (
                                        a["antibiotic"],
                                        a["mic"],
                                        a["sensitivity"],
                                        a.get("cutoff", ""),
                                        a.get("diametre", ""),
                                    )
                                )
                            else:
                                multiple_clarilab.append(
                                    (a["antibiotic"], a["mic"], a["sensitivity"])
                                )
                    elif values and len(values[0].get("multiple_clarilab", [])):
                        multiple_clarilab = values[0]["multiple_clarilab"]

                    if multiple_clarilab:
                        # BEGIN INCANTATION VAUDOU

                        # Spéficique Clarilab
                        # AECO~ATB ESCCOL~L/A_VAM\8\I^ANTIBIO\CMI\SENSI^...
                        # BACT+~Culture~L/#\3C\G_STAAUR\#\#\#\#\#\#\#\#\#\#\#
                        assert len(values) == 1, values
                        data = self.DEFAULT_SEPARATORS[1].join(
                            [
                                values[0]["value"],
                                values[0]["label"],
                                values[0]["codification"],
                            ]
                        )
                        data += "/"
                        lines = []
                        # Attention certaines données peuvent etre des INT.
                        for m in multiple_clarilab:
                            if isinstance(m, (list, tuple)):
                                lines.append(
                                    self.DEFAULT_SEPARATORS[3].join(list(map(str, m)))
                                )
                            else:
                                lines.append(str(m))
                        # ATTENTION: C'est '^' en dur dans le code de clarilab
                        # (sgdl/hprim/resultat.py)
                        data += "^".join(lines)

                        seg_r = do_seg_r()
                        seg_r.set_raw("resu", data)

                        # END INCANTATION VAUDOU
                    elif embended_file:
                        seg_r = do_seg_r()
                        seg_r.set_raw("type_resu", embended_file[2])
                        seg_r.set_raw("resu", embended_file[4])
                    else:
                        seg_r = do_seg_r()
                        for value in values:
                            if fic_file and self.hprim_version == "H2.5":
                                resu = seg_r.resu.add()
                                resu.code1 = worklist.sender_id
                                resu.label1 = value["value"]
                                resu.codification1 = value["value"][-3:].upper()
                            else:
                                resu = seg_r.resu.add()
                                resu.code1 = value["value"]
                                resu.label1 = value.get("label")
                                resu.codification1 = value.get("codification")

                    if (
                        seg_r is not None
                        and result.type == "GC"
                        and all([not v.get("label") for v in values])
                    ):
                        # Graphe à pas fixe de type y^y^y Dans
                        # ce cas il faut renseigner les champs
                        # pas et origine
                        seg_r.gc_pas = "1"
                        seg_r.gc_origine = "0"

                    for comment in result.comments:
                        if seg_r is not None:
                            seg_r.commentaires.add(origine="L", texte=[comment])

        if worklist.context == "result":
            msg = self.clean_result_msg(msg)

        if msg.patients:
            return msg

    def clean_result_msg(self, msg):
        # Ne garde que les segment OBR qui ont des OBX associés
        for seg_p in msg.patients:
            for seg_o in seg_p.orders[:]:
                if not seg_o.results:
                    seg_p.orders.remove(seg_o)

        for seg_p in msg.patients[:]:
            if not seg_p.orders:
                msg.patients.remove(seg_p)

        return msg

    def ignore_seg_p(self, seg_p):
        for seg_obr in seg_p.orders:
            for seg_obx in seg_obr.results:
                if seg_obx.date is None and seg_obx._raw_date:
                    logger.error(
                        "Date OBX mal formée '%s' on ignore le segment P",
                        seg_obx._raw_date,
                    )
                    return True
        return False

    def add_result(self, worklist, sample, seg_obx):
        if worklist.context == "result" and seg_obx.stat_resu == "X":
            return

        if worklist.context == "order" and seg_obx.date is not None:
            anterior = True
        else:
            anterior = False

        if self.hprim_repetiteur_commentaire:
            sep = {"TX": "^"}.get(seg_obx.type_resu, ";")
        else:
            sep = {"TX": "\n"}.get(seg_obx.type_resu, ";")

        if seg_obx.type_resu == "FIC":
            # 10.6.1 -> Nom de l'application éméttrice
            # 10.6.2 -> Identification du fichier
            # 10.6.3 -> Type de fichier
            value = sep.join([r.label1 for r in seg_obx.resu])
            label = seg_obx.code2
        else:
            value = sep.join([r.code1 for r in seg_obx.resu])
            label = sep.join([r.label1 for r in seg_obx.resu])

        flags = []
        if self.option_comments_in_flags:
            for seg_c in seg_obx.commentaires:
                flags.extend(seg_c.texte)
        if seg_obx.type_resu == "FIC" and seg_obx.sous_test:
            code = "%s_%s" % (seg_obx.code, seg_obx.sous_test)
        else:
            code = seg_obx.code
        if seg_obx.droit_acces:
            instrument = seg_obx.droit_acces
        else:
            instrument = ""
        if self.option_hprim_integre_rack and code == "RACK":
            sample.rack = value
        result_kwargs = dict(
            type=seg_obx.type_resu,
            label=seg_obx.libelle,
            codification=seg_obx.codification,
            alt_code=seg_obx.code2,
            alt_label=seg_obx.libelle2,
            alt_codification=seg_obx.codification2,
            sub_code=seg_obx.sous_test,
            date=seg_obx.date,
            flags=flags,
            anterior=anterior,
            unit=seg_obx.unite.code1,
            unit_label=seg_obx.unite.label1,
            unit_codification=seg_obx.unite.codification1,
            alt_unit=seg_obx.unite.code2,
            alt_unit_label=seg_obx.unite.label2,
            alt_unit_codification=seg_obx.unite.codification2,
            instrument=instrument,
        )
        if self.hprim_concatener_resultats_multiples:
            # Traitement des résultats provenant d'une connexion Hprim ou
            # BacExpress de type automate
            result = sample.results.get_or_create(code=code)
            if result.value:
                result_kwargs["value_label"] = ";".join((result.value_label, label))
                result_kwargs["value"] = ";".join((result.value, value))
            else:
                result_kwargs["value"] = value
                result_kwargs["value_label"] = label
            result.update(**result_kwargs)
        else:
            # Traitement des résultats provenant d'une connexion de type SIL
            # avec les résultats accueils ou les antériorités
            result = sample.results.get_or_create(
                code=code, value=value, value_label=label, **result_kwargs
            )

        # LC: Doing this as a if prevents "normality"
        # from apearing all the time inside the worklist dictionary
        if seg_obx.normale:
            result.normality = seg_obx.normale
        if seg_obx.valideur:
            result.validator = seg_obx.valideur
        if self.hprim_send_validator_code_and_name:
            result.validator_code = seg_obx.valideur_code
            result.validator = seg_obx.valideur_nom
        if anterior:
            result.request_id = seg_obx.no_demande
            result.request_date = seg_obx.date_demande
        self._decode_secteurtech(seg_obx, result)

    def decode_message(self, message):
        if message.type_message not in ("ORU", "ORM"):
            raise NotImplementedError("type message %s" % (message.type_message,))
        worklist = Worklist(
            context="result" if message.type_message == "ORU" else "order",
            sender_id=message.emetteur.code,
            receiver_id=message.recepteur.code,
        )
        for seg_p in message.patients:
            if self.ignore_seg_p(seg_p):
                continue
            for seg_obr in seg_p.orders:
                sample = worklist.get_or_create(
                    sample_id=seg_obr.tube_demandeur,
                    request_id=seg_obr.no_demandeur,
                    exe_sample_id=seg_obr.tube_executant,
                    exe_request_id=seg_obr.no_executant,
                    patient_id=seg_p.patient_demandeur
                    or seg_p.patient_executant
                    or seg_obr.no_demandeur,
                )
                if seg_p.telephone == [["", "T"]]:
                    # À telephoner clarilab...
                    phone = ""
                else:
                    phone = ";".join([p for p in seg_p.telephone])
                # Date de la demande
                if self.hprim_date_prel:
                    request_date = seg_obr.date_prel or message.date
                else:
                    request_date = (
                        seg_obr.date_demande or seg_obr.date_prel or message.date
                    )
                sample.update(
                    alternate_patient_id=seg_p.no_adm_cli,
                    last_name=seg_p.nom.nom,
                    first_name=seg_p.nom.prenom,
                    title=seg_p.nom.civilite,
                    maiden_name=seg_p.nomjf,
                    birthdate=seg_p.date_nai,
                    birth_row=seg_p.rang_naissance,
                    sex=seg_p.sexe if seg_p.sexe in ("M", "MS", "F", "FS") else "U",
                    request_date=request_date,
                    type="patient",
                    diagnostic=";".join([e.code1 for e in seg_p.diagnostic_patient]),
                    location=seg_p.localisation.corr,
                    room=seg_p.localisation.chambre,
                    bed=seg_p.localisation.lit,
                    collection_date=seg_obr.date_prel,
                    collection_end_date=seg_obr.date_fin_prel,
                    admission_date=seg_p.date_entree,
                    admission_end_date=seg_p.date_sortie,
                    height=seg_p.taille,
                    phone=phone,
                    transcriptionist=seg_obr.ope_saisie.code,
                    city=seg_p.adresse.ville,
                    postal_code=seg_p.adresse.code_postal,
                    country=seg_p.adresse.pays,
                    address="~".join(
                        [
                            li
                            for li in [seg_p.adresse.ligne1, seg_p.adresse.ligne2]
                            if li
                        ]
                    ),
                    admission_status=seg_p.stat_adm,
                    deathdate=seg_p.date_deces,
                    physiologie=self.get_physiologie(seg_p),
                )
                if message.version == "H2.5":
                    for identity in seg_p.birth_identity:
                        further_names = []
                        if identity.further_names:
                            for name in identity.further_names.split(" "):
                                further_names.append(name)
                        sample.update(
                            last_name=identity.last_name,
                            first_name=identity.first_name,
                            further_names=further_names,
                        )
                    sample.update(
                        used_last_name=seg_p.nom.nom, used_first_name=seg_p.nom.prenom
                    )
                    for ins in seg_p.ins_number:
                        sample.update(
                            ins_number=ins.identifier,
                            ins_type=ins.type,
                            insi_query_date=ins.call_date,
                            ins_effective_date=ins.effective_date,
                            ins_end_date=ins.end_date,
                            oid=ins.oid,
                        )
                    if seg_p.birth_place:
                        for place in seg_p.birth_place:
                            sample.update(
                                birthplace=place.name,
                                birthplace_code=place.insee_code,
                                birthplace_country=place.country,
                                birthplace_postal_code=place.postal_code,
                            )
                    sample.update(
                        identity_obtaining_mode=seg_p.identity_obtaining_mode,
                        identity_document=seg_p.identity_document,
                        birthdate_modified=seg_p.birthdate_modified,
                        identity_status=seg_p.identity_status,
                        identity_attribut=seg_p.identity_attribut,
                    )
                    if seg_p.lunar_date:
                        for date in seg_p.lunar_date:
                            lunar_date = "{}{}{}".format(
                                date.year, date.month, date.day
                            )
                        sample.update(lunar_date=lunar_date)
                if self.option_identification_demandeur:
                    if self.option_prefixe_code_executant_labo:
                        sample.request_info = "{}{}".format(
                            self.option_prefixe_code_executant_labo, seg_obr.lib1_dem
                        )
                    else:
                        sample.request_info = seg_obr.lib1_dem
                if seg_p.poids:
                    sample.weight = str(seg_p.poids)
                for service in seg_obr.get_services():
                    if (
                        seg_obr.prescripteur
                        and seg_obr.prescripteur[0].code_service
                        and message.version == "H2.4"
                    ):
                        continue
                    else:
                        sample.services.get_or_create(**service)
                for type, segments in (
                    ("patient", seg_p.commentaires),
                    ("sample", seg_obr.commentaires),
                ):
                    for seg_c in segments:
                        for item in seg_c.texte:
                            sample.comments.get_or_create(type=type, text=item)
                for item in seg_obr.get_comments(self.option_identification_demandeur):
                    sample.comments.get_or_create(type="sample", text=item)
                for item in [
                    seg_p.traite_lc1,
                    seg_p.traite_lc2,
                    seg_p.comment_dem1,
                    seg_p.comment_dem2,
                ]:
                    if item:
                        sample.comments.get_or_create(type="patient", text=item)
                for item in seg_p.traitement_medical:
                    sample.medications.get_or_create(item)
                for med in seg_p.medecins:
                    if med.code not in [p.code for p in sample.physicians]:
                        sample.physicians.get_or_create(
                            code=med.code,
                            last_name=med.nom.nom,
                            first_name=med.nom.prenom,
                            type="attending",
                            codification=med.type,
                            title=med.nom.civilite,
                            alias=med.nom.sobriquet,
                        )
                self._add_custom_info_patient(sample, seg_p)
                self.get_medecin_prescripteur(sample, seg_obr)

                if seg_obr.preleveur.code:
                    sample.physicians.get_or_create(
                        type="sampler",
                        code=seg_obr.preleveur.code,
                        last_name=seg_obr.preleveur.nom.nom,
                        first_name=seg_obr.preleveur.nom.prenom,
                        codification=seg_obr.preleveur.type,
                        title=seg_obr.preleveur.nom.civilite,
                        alias=seg_obr.preleveur.nom.sobriquet,
                    )
                if seg_obr.nature.type_tube.code:
                    nature = seg_obr.nature.type_tube
                    sample.nature.code = nature.code
                    sample.nature.label = nature.label
                    sample.nature.codification = nature.codification
                    sample.nature.additive = seg_obr.nature.additif

                if worklist.context == "order":
                    combined_orders_results = []
                    for p in seg_obr.prescription:
                        if p.code1 in [x["code"] for x in combined_orders_results]:
                            logger.debug(
                                "code %s already present: do not add.", p.code1
                            )
                            continue
                        combined_orders_results.append(
                            {
                                "code": p.code1,
                                "label": p.label1,
                                "codification": p.codification1,
                                "alt_code": p.code2,
                                "alt_label": p.label2,
                                "alt_codification": p.codification2,
                                "accueil": False,
                            }
                        )
                    for r in [r for r in seg_obr.results if r.date is None]:
                        if r.code in [x["code"] for x in combined_orders_results]:
                            logger.debug("code %s already present: do not add.", r.code)
                            continue
                        combined_orders_results.append(
                            {
                                "code": r.code,
                                "label": r.label if hasattr(r, "label") else None,
                                "codification": (
                                    r.codification
                                    if hasattr(r, "codification")
                                    else None
                                ),
                                "alt_code": (
                                    r.alt_code if hasattr(r, "alt_code") else None
                                ),
                                "alt_label": (
                                    r.alt_label if hasattr(r, "alt_label") else None
                                ),
                                "alt_codification": (
                                    r.alt_codification
                                    if hasattr(r, "alt_codification")
                                    else None
                                ),
                                "accueil": True,
                            }
                        )
                    for orderd in combined_orders_results:
                        sampleorder = sample.orders.get_or_create(
                            code=orderd["code"],
                            priority=seg_obr.priorite if seg_obr.priorite else "R",
                            action=seg_obr.get_action(),
                        )
                        if orderd.get("label"):
                            sampleorder.label = orderd["label"]
                        if orderd.get("alt_code"):
                            sampleorder.alt_code = orderd["alt_code"]
                        if orderd.get("alt_label"):
                            sampleorder.alt_label = orderd["alt_label"]
                        if orderd.get("accueil"):
                            sampleorder.accueil = orderd["accueil"]
                self.decode_obr(seg_obr, worklist, sample)
        return worklist

    def decode_obr(self, seg_obr, worklist, sample):
        for seg_obx in seg_obr.results:
            self.add_result(worklist, sample, seg_obx)

    def _decode_secteurtech(self, seg_obx, result):
        if seg_obx.secteur_tech.libelle and not result.chapter:
            result.chapter.code = seg_obx.secteur_tech.libelle
            if seg_obx.secteur_tech.rang:
                result.chapter.position = seg_obx.secteur_tech.rang
            if seg_obx.secteur_tech.code_batterie:
                result.battery.code = seg_obx.secteur_tech.code_batterie
            if seg_obx.secteur_tech.rang_batterie:
                result.battery.chapter_position = seg_obx.secteur_tech.rang_batterie
            if seg_obx.secteur_tech.rang_test:
                result.battery.position = seg_obx.secteur_tech.rang_test

    def get_physiologie(self, seg_p):
        return ""

    def get_medecin_prescripteur(self, sample, seg_obr):
        for i, prescripteur in enumerate(seg_obr.prescripteur):
            if prescripteur.code_medecin and prescripteur.code_medecin not in [
                p.code for p in sample.physicians
            ]:
                sample.physicians.get_or_create(
                    code=prescripteur.code_medecin,
                    last_name=prescripteur.medecin.nom,
                    first_name=prescripteur.medecin.prenom,
                    codification=prescripteur.type_medecin,
                    type="prescriber",
                    phone=";".join(seg_obr.tel_prescrip) if not i else None,
                )

    def _add_custom_info_patient(self, segment, patient):
        pass
